from typing import Optional

from pyinaturalist.constants import IntOrStr, MultiIntOrStr
from pyinaturalist.controllers import BaseController
from pyinaturalist.converters import ensure_list
from pyinaturalist.models import User
from pyinaturalist.paginator import IDPaginator, Paginator
from pyinaturalist.v1 import get_current_user, get_user_by_id, get_users_autocomplete


class UserController(BaseController):
    """:fa:`user` Controller for User requests"""

    def __call__(self, user_id: IntOrStr, **kwargs) -> Optional[User]:
        """Get a single user by ID

        Example:
            >>> user = client.users(1)

        Args:
            user_id: A single user ID
        """
        return self.from_ids(user_id, **kwargs).one()

    def from_ids(self, user_ids: MultiIntOrStr, **params) -> Paginator[User]:
        """Get users by ID

        Example:
            Get a user by ID:

            >>> user = client.users.from_id(1).one()

            Get multiple users by ID:

            >>> users = client.users.from_id([1,2]).all()

        Args:
            user_ids: One or more user IDs
        """
        return self.client.paginate(
            get_user_by_id, User, cls=IDPaginator, ids=ensure_list(user_ids), **params
        )

    def autocomplete(
        self, q: Optional[str] = None, project_id: Optional[int] = None, **params
    ) -> Paginator[User]:
        """Given a query string, return users with names or logins starting with the search term

        .. rubric:: Notes

        * API reference: :v1:`GET /users/autocomplete <Users/get_users_autocomplete>`

        Example:
            >>> client.users.autocomplete(q='my_userna')

        Args:
            q: Search query
            project_id: Only show users who are members of this project
        """
        return self.client.paginate(
            get_users_autocomplete, User, q=q, project_id=project_id, **params
        )

    def me(self, **params) -> User:
        """Get your own user profile

        .. rubric:: Notes

        * :fa:`lock` :ref:`Requires authentication <auth>`
        * API reference: :v1:`GET /users/me <Users/get_users_me>`

        Example:
            >>> client.users.me()
        """
        response = self.client.request(get_current_user, auth=True, **params)
        return User.from_json(response)



from datetime import datetime
from typing import List

from pyinaturalist.constants import INAT_BASE_URL, JsonResponse, TableRow
from pyinaturalist.models import (
    BaseModel,
    BaseModelCollection,
    datetime_now_field,
    define_model,
    define_model_collection,
    field,
)


@define_model
class User(BaseModel):
    """:fa:`user` An iNaturalist user, based on the schema of :v1:`GET /users/{id} <Users/get_users_id>`"""

    activity_count: int = field(
        default=0,
        doc='Combined user activity including observations, identifications, and journal posts',
    )
    annotated_observations_count: int = field(
        default=0, doc='Number of observations the user has annotated'
    )
    created_at: datetime = datetime_now_field(doc='Date and time the user was registered')
    icon: str = field(default=None, doc='URL for small user icon')
    icon_url: str = field(default=None, doc='URL for medium user icon')
    identifications_count: int = field(default=0, doc='Number of identifications the user has made')
    journal_posts_count: int = field(default=0, doc='Number of journal posts the user has made')
    login: str = field(default=None, doc='User login/username')
    name: str = field(default=None, doc='User real name or display name')
    observations_count: int = field(default=0, doc='Number of observations the user has made')
    orcid: str = field(default=None, doc='ORCID iD')
    roles: List[str] = field(factory=list, doc='User roles on inaturalist.org')
    site_id: int = field(
        default=None, doc='Site ID for iNaturalist network members, or ``1`` for inaturalist.org'
    )
    species_count: int = field(default=0, doc='Number of unique species the user has observed')

    # Unused attributes
    # login_autocomplete: str = field(default=None)
    # login_exact: str = field(default=None)
    # name_autocomplete: str = field(default=None)
    # spam: bool = field(default=None)
    # suspended: bool = field(default=None)
    # universal_search_rank: int = field(default=None)

    @property
    def username(self) -> str:
        """Alias of ``login``"""
        return self.login

    @property
    def display_name(self) -> str:
        """Alias of ``name``"""
        return self.name

    @property
    def url(self) -> str:
        """User info URL on iNaturalist.org"""
        return f'{INAT_BASE_URL}/users/{self.id}'

    @property
    def _row(self) -> TableRow:
        return {
            'ID': self.id,
            'Username': self.username,
            'Display name': self.display_name,
            'Observations': self.observations_count,
            'Identifications': self.identifications_count,
        }

    @property
    def _str_attrs(self) -> List[str]:
        return ['id', 'login', 'name']


@define_model
class UserCount(User):
    """:fa:`user` An iNaturalist user, with an associated count of filtered IDs or observations"""

    count: int = field(default=0, doc='Filtered count for the user')
    observation_count: int = field(default=0, doc="Filtered count for the user's observations")
    species_count: int = field(
        default=0, doc="Filtered count for the user's unique species observed"
    )

    @classmethod
    def from_json(cls, value: JsonResponse, **kwargs) -> 'UserCount':
        """Flatten out count + user fields into a single-level dict before initializing"""
        if 'results' in value:
            value = value['results']
        if 'user' in value:
            value = value.copy()
            value.update(value.pop('user'))
        if 'observation_count' in value and 'count' not in value:
            value['count'] = value['observation_count']
        return super(UserCount, cls).from_json(value)

    @property
    def _row(self) -> TableRow:
        return {
            'ID': self.id,
            'Username': self.username,
            'Display name': self.display_name,
            'Count': self.count,
        }

    def __str__(self) -> str:
        return super().__str__() + f': {self.count}'


@define_model_collection
class UserCounts(BaseModelCollection):
    """:fa:`user` :fa:`list` A collection of users with an associated counts"""

    data: List[UserCount] = field(factory=list, converter=UserCount.from_json_list)



"""Classes to handle pagination of API requests"""

from asyncio import AbstractEventLoop, get_running_loop
from collections import deque
from concurrent.futures import ThreadPoolExecutor
from logging import getLogger
from math import ceil
from typing import (
    TYPE_CHECKING,
    AsyncIterable,
    AsyncIterator,
    Callable,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Type,
)

from requests import Response

from pyinaturalist.constants import (
    EXPORT_URL,
    LARGE_REQUEST_WARNING,
    PER_PAGE_RESULTS,
    REQUESTS_PER_MINUTE,
    IntOrStr,
    JsonResponse,
    RequestParams,
    ResponseResult,
)
from pyinaturalist.models import T

logger = getLogger(__name__)


# TODO: Add per-endpoint 'max_per_page' parameter to use with Paginator.all()
class Paginator(Iterable, AsyncIterable, Generic[T]):
    """Class to handle pagination of API requests, with async support

    Args:
        request_function: API request function to paginate
        model: Model class to use for results
        limit: Maximum number of total results to fetch
        per_page: Maximum number of results to fetch per page
        loop: An event loop to use to run any executors used for async iteration
        kwargs: Original request parameters
    """

    def __init__(
        self,
        request_function: Callable,
        model: Type[T],
        *request_args,
        limit: Optional[int] = None,
        per_page: Optional[int] = None,
        loop: Optional[AbstractEventLoop] = None,
        **request_kwargs,
    ):
        self.request_function = request_function
        self.request_args = request_args
        self.request_kwargs = {k: v for k, v in request_kwargs.items() if v is not None}
        self.request_kwargs.pop('page', None)

        self.exhausted = False
        self.loop = loop
        self.model = model
        self.per_page = per_page or PER_PAGE_RESULTS
        self.page = 1
        self.results_fetched = 0
        self.total_limit = limit
        self.total_results: Optional[int] = None

        if request_function is not None:
            log_kwargs = {
                k: v for k, v in self.request_kwargs.items() if k not in ['session', 'access_token']
            }
            logger.debug(
                f'Prepared paginated request: {self.request_function.__name__}'
                f'(args={self.request_args}, kwargs={log_kwargs})'
            )

    async def __aiter__(self) -> AsyncIterator[T]:
        """Iterate over paginated results, with non-blocking requests sent from a separate thread"""
        loop = self.loop or get_running_loop()
        with ThreadPoolExecutor(max_workers=1) as executor:
            while not self.exhausted:
                for result in await loop.run_in_executor(executor, self.next_page):
                    yield result

    def __iter__(self) -> Iterator[T]:
        """Iterate over paginated results"""
        while not self.exhausted:
            for result in self.next_page():
                yield result

    async def async_all(self) -> List[T]:  # Better name TBD?
        """Get all results in a single list (non-blocking)"""
        return [result async for result in self]

    def all(self) -> List[T]:
        """Get all results in a single list"""
        return list(self)

    def limit(self, limit: int) -> List[T]:
        """Get at most ``limit`` results"""
        self.total_limit = limit
        return list(self)

    def one(self) -> Optional[T]:
        """Get the first result from the query"""
        self.per_page = 1
        results = self.next_page()
        return results[0] if results else None

    def count(self) -> int:
        """Get the total number of results for this query, without fetching response data.

        Returns:
            Either the total number of results, if the endpoint provides pagination info, or ``-1``
        """
        if self.total_results is None:
            kwargs = {**self.request_kwargs, 'per_page': 0}
            response = self.request_function(*self.request_args, **kwargs)
            if isinstance(response, Response):
                response = response.json()
            self.total_results = int(response['total_results'])
        return self.total_results

    def next_page(self) -> List[T]:
        """Get the next page of results, as model objects"""
        return self.model.from_json_list(self._next_page())

    def _next_page(self) -> List[ResponseResult]:
        """Get the next page of results, as raw JSON"""
        if self.exhausted:
            return []

        # If a limit is specified, avoid fetching more results than needed
        if self.total_limit and self.results_fetched + self.per_page > self.total_limit:
            self.per_page = self.total_limit - self.results_fetched

        # Fetch results; handle response object or dict
        response = self.request_function(
            *self.request_args,
            **self.request_kwargs,
            **self._get_pagination_kwargs(),
        )
        if isinstance(response, Response):
            response = response.json()
        results = response.get('results', response)

        # Note: For id-based pagination, only the first page's 'total_results' is accurate
        if self.total_results is None:
            self.total_results = response.get('total_results', len(results))
        self.results_fetched += len(results)

        # If this is the first of multiple requests, log the estimated time and number of requests
        if not self._check_exhausted(results):
            if self.page == 1:
                self._estimate()
            self.page += 1

        return results

    # The following two methods may be overridden by subclasses for different pagination methods
    def _get_pagination_kwargs(self) -> RequestParams:
        """Get any extra request parameters needed for pagination"""
        return {'page': self.page, 'per_page': self.per_page}

    def _check_exhausted(self, page_results: Optional[List] = None) -> bool:
        """Check all conditions that indicate no more results are available.
        (relevant conditions and error cases vary by API endpoint)
        """
        n_results = len(page_results or [])
        self.exhausted = any(
            [
                self.exhausted,
                (self.total_limit and self.results_fetched >= self.total_limit),
                (self.total_results and self.results_fetched >= self.total_results),
                (self.per_page and n_results < self.per_page),
                n_results == 0,
            ]
        )
        return self.exhausted

    def _deduplicate(self, results) -> List[T]:
        """Deduplicate results by ID"""
        unique_results = {result.id: result for result in results}
        self.total_results = len(unique_results)
        return list(unique_results.values())

    def _estimate(self):
        """Log the estimated total number of requests and rate-limiting delay, and show a warning if
        the request is too large
        """
        total_requests = ceil(self.total_results / self.per_page)
        est_delay = ceil((total_requests / REQUESTS_PER_MINUTE) * 60) - 1
        logger.info(
            f'This query will fetch {self.total_results} results in {total_requests} requests. '
            f'Estimated total request time: {est_delay} seconds'
        )

        if self.total_results > LARGE_REQUEST_WARNING:
            logger.warning(
                'This request is larger than recommended for API usage. For bulk requests, consider '
                f'using the iNat export tool instead: {EXPORT_URL}'
            )

    def __str__(self) -> str:
        return (
            f'{self.__class__.__name__}(request_function={self.request_function.__name__}, '
            f'fetched={self.results_fetched}/{self.total_results or "unknown"})'
        )


class IDRangePaginator(Paginator):
    """Paginate by a range of IDs instead of standard pagination parameters

    .. note::
        Note on pagination by ID, from the iNaturalist documentation:

        *The large size of the observations index prevents us from supporting the page parameter
        when retrieving records from large result sets. If you need to retrieve large numbers of
        records, use the ``per_page`` and ``id_above`` or ``id_below`` parameters instead.*

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.id_above: Optional[int] = None

    def _get_pagination_kwargs(self):
        return {
            'id_above': self.id_above,
            'per_page': self.per_page,
            'order_by': 'id',
            'order': 'asc',
        }

    def _next_page(self) -> List[ResponseResult]:
        # Update params for next request, if there are more results
        results = super()._next_page()
        if not self.exhausted:
            self.id_above = results[-1]['id']
        return results


class IDPaginator(Paginator):
    """Paginator for ID-based endpoints that only accept a limited number of IDs per request"""

    def __init__(
        self, *args, ids: Optional[Iterable[IntOrStr]] = None, ids_per_request: int = 1, **kwargs
    ):
        super().__init__(*args, **kwargs)
        if ids_per_request == 1:
            self.id_batches = deque(ids or [])
        else:
            self.id_batches = deque(list(_chunkify(ids, ids_per_request)))  # type: ignore
        self.total_results = len(ids)  # type: ignore

    def _next_page(self) -> List[ResponseResult]:
        """Get the next record by ID"""
        try:
            next_ids = self.id_batches.popleft()
        except IndexError:
            self.exhausted = True
            return []

        response = self.request_function(next_ids, *self.request_args, **self.request_kwargs)
        if isinstance(response, Response):
            response = response.json()

        results = response['results'] if 'results' in response else [response]
        self.page += 1
        self.results_fetched += len(results)
        return results


def _chunkify(iterable: Iterable, max_size: int) -> Iterator[List]:
    """Split an iterable into chunks of a max size"""
    iterable = list(iterable)
    for index in range(0, len(iterable), max_size):
        yield iterable[index : index + max_size]


class AutocompletePaginator(Paginator):
    """Paginator that attempts to get as many results as possible from an autocomplete endpoint.
    This is necessary for some problematic queries for which there are many matches but not ranked
    with the desired match(es) first.

    This works based on different rankings being returned for order_by=area. No other fields can be
    sorted on, and direction can't be specified, but this can at least provide a few additional
    results beyond the limit of 20.

    All results will be de-duplicated and returned as a single page. This may potentially be applied
    to other autocomplete endpoints, but so far is only needed for places.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request_kwargs.pop('order_by', None)

    def all(self) -> List[T]:
        """Get all results in a single de-duplicated list"""
        return self._deduplicate(list(self))

    def _check_exhausted(self, page_results: Optional[List] = None):
        """Also check for the odd case in which we get more results than requested"""
        self.exhausted = any(
            [
                super()._check_exhausted(page_results),
                (self.per_page and self.results_fetched > self.per_page),
            ]
        )

    def _get_pagination_kwargs(self):
        kwargs = super()._get_pagination_kwargs()
        if not self.exhausted and self.page >= 1:
            kwargs['order_by'] = 'area'
        return kwargs


if TYPE_CHECKING:
    MixinBase = Paginator
else:
    MixinBase = object


class JsonPaginatorMixin(MixinBase):
    """Paginator mixin that returns raw response dicts instead of model objects"""

    def __init__(
        self,
        request_function: Callable,
        *request_args,
        **kwargs,
    ):
        super().__init__(request_function, None, *request_args, **kwargs)  # type: ignore

    def __iter__(self) -> Iterator[ResponseResult]:
        """Iterate over paginated results, and skip conversion to model objects"""
        while not self.exhausted:
            for result in self._next_page():
                yield result

    def all(self) -> JsonResponse:  # type: ignore
        results = super().all()
        return {
            'results': results,
            'total_results': len(results),
        }

    def _deduplicate(self, results) -> List[T]:
        """Deduplicate results by ID"""
        unique_results = {result['id']: result for result in results}
        self.total_results = len(unique_results)
        return list(unique_results.values())


class JsonPaginator(JsonPaginatorMixin, Paginator):
    pass


class JsonIDRangePaginator(JsonPaginatorMixin, IDRangePaginator):
    pass


def paginate_all(request_function: Callable, *args, method: str = 'page', **kwargs) -> JsonResponse:
    """Get all pages of a multi-page request. Explicit pagination parameters will be overridden.

    Returns:
        Response dict containing combined results, in the same format as ``api_func``
    """
    paginator = JsonIDRangePaginator if method == 'id' else JsonPaginator
    return paginator(request_function, *args, **kwargs).all()


class WrapperPaginator(Paginator):
    """Paginator class that wraps results that have already been fetched."""

    def __init__(self, results: List[T], *args, **kwargs):
        super().__init__(None, None, *args, **kwargs)  # type: ignore  # no request_function
        self.results = results
        self.results_fetched = self.total_results = len(results)

    def count(self) -> int:
        return len(self.results)

    def next_page(self):
        self.exhausted = True
        return self.results



from typing import List

from pyinaturalist.constants import DATETIME_SHORT_FORMAT, DateTime, TableRow
from pyinaturalist.models import BaseModel, LazyProperty, User, datetime_field, define_model, field


@define_model
class Message(BaseModel):
    """:fa:`user` A message from the user's inbox, based on the schema of
    :v1:`GET /messages <Messages/get_messages>`
    """

    user_id: int = field(default=None, doc='Corresponding user ID')
    thread_id: int = field(default=None, doc='Message thread ID')
    subject: str = field(default=None, doc='Message subject')
    body: str = field(default=None, doc='Message body')
    read_at: DateTime = datetime_field(doc='When the message was read')
    created_at: DateTime = datetime_field(doc='When the message was sent')
    updated_at: DateTime = datetime_field(doc='When the message was last edited')
    thread_flags: List[str] = field(factory=list)
    thread_messages_count: int = field(default=None, doc='Number of messages in the thread')

    from_user: property = LazyProperty(User.from_json, type=User, doc='Message sender')
    to_user: property = LazyProperty(User.from_json, type=User, doc='Message recipient')

    @property
    def truncated_body(self):
        """Comment text, truncated"""
        truncated_body = self.body.replace('\n', ' ').strip()
        if len(truncated_body) > 50:
            truncated_body = truncated_body[:47].strip() + '...'
        return truncated_body

    @property
    def _row(self) -> TableRow:
        return {
            'ID': self.id,
            'Date': self.created_at,
            'To': self.to_user.login,
            'From': self.from_user.login,
            'Subject': self.subject,
        }

    @property
    def _str_attrs(self) -> List[str]:
        return ['id', 'created_at', 'self.to_user', 'self.from_user', 'subject', 'truncated_body']

    def __str__(self) -> str:
        return (
            f'[{self.id}] Sent {self.created_at.strftime(DATETIME_SHORT_FORMAT)} '
            f'from {self.from_user.login} to {self.to_user.login}: {self.subject}'
        )



"""Utilities for modifying function signatures using ``python-forge``"""

# ruff: noqa: E501
from functools import partial
from inspect import Parameter, ismethod, signature
from logging import getLogger
from typing import Callable, Dict, Iterable, List, Optional, Type

from requests import Session

from pyinaturalist.constants import TemplateFunction
from pyinaturalist.converters import ensure_list
from pyinaturalist.docs import copy_annotations, copy_docstrings, forge

AUTOMETHOD_INIT = '.. automethod:: __init__'
CONTROLLER_EXCLUDE_PARAMS = ['dry_run', 'session', 'page', 'per_page', 'order', 'count_only']

logger = getLogger(__name__)


def copy_doc_signature(
    *template_functions: TemplateFunction,
    add_common_args: bool = False,
    description_only: bool = False,
    include_sections: Optional[Iterable[str]] = None,
    include_return_annotation: bool = True,
    exclude_args: Optional[Iterable[str]] = None,
) -> Callable:
    """Document a function with docstrings, function signatures, and type annotations from
    one or more template functions.

    If used with other decorators, this should go first (e.g., last in the call order).

    Example:
        >>> # 1. Template function with individual request params + docs
        >>> def get_foo_template(arg_1: str = None, arg_2: bool = False):
        >>>     '''Args:
        >>>     arg_1: Example request parameter 1
        >>>     arg_2: Example request parameter 2
        >>>     '''

        >>> # 2. Decorated API Endpoint function with generic (variadic) keyword args
        >>> @document_request_params(get_foo_template)
        >>> def get_foo(**kwargs) -> List:
            ''' Get Foo resource '''

        >>> # 3. Modified function signature + docstring
        >>> help(get_foo)
        '''
        Help on function get_foo:
        get_foo(arg_1: str = None, arg_2: bool = False) -> List
        Get Foo resource
        Args:
            arg_1: Example request parameter 1
            arg_2: Example request parameter 2
        '''

    Args:
        template_functions: Template functions containing docstrings and params to apply to the
            wrapped function
        add_common_args: Add additional keyword arguments common to most functions
        description_only: Only add the function description
        include_sections: Docstring sections to include; if not specified, all sections will be included
        include_return_annotation: Copy the return type annotation from the template function(s)
        exclude_args: Arguments to exclude from the new docstring
    """
    if add_common_args:
        template_functions += (_dry_run, _session)
    if description_only:
        include_sections = ['Description']

    def wrapper(func):
        try:
            if not description_only:
                func = copy_annotations(func, template_functions, include_return_annotation)
            func = copy_docstrings(func, template_functions, include_sections, exclude_args)
            if not description_only:
                func = copy_signatures(func, template_functions, exclude_args)
        # If for any reason one of these steps fail, just log the error and return the original function
        except Exception:
            logger.exception(f'Failed to modify {func.__name__}')
        return func

    return wrapper


# Alias specifically for basic request functions
document_request_params = partial(copy_doc_signature, add_common_args=True)

# Aliases specifically for controller functions
document_controller_params = partial(
    copy_doc_signature,
    include_sections=['Description', 'Args'],
    include_return_annotation=False,
    exclude_args=CONTROLLER_EXCLUDE_PARAMS,
)
document_controller_description = partial(
    copy_doc_signature,
    description_only=True,
)


def document_common_args(func):
    """Just add common args to a function's docstring without modifying the signature"""
    template_functions = (_dry_run, _session)
    func = copy_annotations(func, template_functions)
    func = copy_docstrings(func, template_functions)
    return func


def extend_init_signature(*template_functions: Callable) -> Callable:
    """A class decorator that behaves like :py:func:`.copy_doc_signature`, but modifies a class
    docstring and its ``__init__`` function signature, and extends them instead of replacing them.
    """

    def wrapper(target_class: Type):
        # Modify init signature + docstring
        revision = copy_doc_signature(*template_functions, target_class.__init__)
        target_class.__init__ = revision(target_class.__init__)

        # Include init docs in class docs
        target_class.__doc__ = target_class.__doc__ or ''
        if AUTOMETHOD_INIT not in target_class.__doc__:
            target_class.__doc__ += f'\n\n    {AUTOMETHOD_INIT}\n'
        return target_class

    return wrapper


def copy_signatures(
    target_function: Callable,
    template_functions: List[TemplateFunction],
    exclude_args: Optional[Iterable[str]] = None,
) -> Callable:
    """A decorator that copies function signatures from one or more template functions to a
    target function.

    Args:
        target_function: Function to modify
        template_functions: Functions containing params to apply to ``target_function``
    """
    # Start with 'self' parameter if this is an instance method
    fparams = {}
    if 'self' in signature(target_function).parameters or ismethod(target_function):
        fparams['self'] = forge.self

    # Add and combine parameters from all template functions, excluding duplicates, self, and *args
    for func in template_functions:
        new_fparams = {
            k: v
            for k, v in forge.copy(func).signature.parameters.items()
            if k != 'self' and v.kind != Parameter.VAR_POSITIONAL
        }
        fparams.update(new_fparams)

    # Manually remove any excluded parameters
    for key in ensure_list(exclude_args):
        fparams.pop(key, None)

    fparams = deduplicate_var_kwargs(fparams)
    revision = forge.sign(*fparams.values())
    return revision(target_function)


def deduplicate_var_kwargs(params: Dict) -> Dict:
    """Add variadic keyword args (e.g., ``**kwargs``) to the end of a function signature, accounting
    for any duplicates.
    """
    # Check for **kwargs by param type instead of by name
    for k, v in params.copy().items():
        if v.kind == Parameter.VAR_KEYWORD:
            params.pop(k)

    # Add **kwargs as the last param
    params.update(forge.kwargs)
    return params


# Templates for common params that are added to every request function by default
# ----------


def _dry_run(dry_run: Optional[bool] = False):
    """Args:
    dry_run: Just log the request instead of sending a real request
    """


def _session(session: Optional[Session] = None):
    """Args:
    session: An existing `Session object <https://docs.python-requests.org/en/latest/user/advanced/>`_ to use instead of creating a new one
    """



from io import BytesIO
from typing import BinaryIO, List, Optional, Tuple

import requests

from pyinaturalist.constants import (
    ALL_LICENSES,
    CC_LICENSES,
    ICON_SIZES,
    ICONIC_TAXA_BASE_URL,
    PHOTO_BASE_URL,
    PHOTO_CC_BASE_URL,
    PHOTO_INFO_BASE_URL,
    PHOTO_SIZES,
    JsonResponse,
    TableRow,
)
from pyinaturalist.converters import format_dimensions, format_license
from pyinaturalist.models import BaseModel, datetime_field, define_model, field


@define_model
class BaseMedia(BaseModel):
    attribution: str = field(default=None, doc='License attribution')
    license_code: str = field(
        default=None,
        converter=format_license,
        options=ALL_LICENSES,
        doc='Creative Commons license code',
    )
    created_at: str = datetime_field(doc='Date the file was added to iNaturalist')
    updated_at: str = datetime_field(doc='Date the file was last updated on iNaturalist')

    @property
    def ext(self) -> str:
        """File extension from URL"""
        return self.url.lower().split('.')[-1].split('?')[0]  # type: ignore

    @property
    def has_cc_license(self) -> bool:
        """Determine if this photo has a Creative Commons license"""
        return self.license_code in CC_LICENSES

    @property
    def _str_attrs(self) -> List[str]:
        return ['id', 'license_code', 'url']


@define_model
class Photo(BaseMedia):
    """:fa:`camera` An observation photo, based on the schema of photos from:

    * `GET /observations <https://api.inaturalist.org/v1/docs/#!/Observations/get_observations>`_
    * `GET /taxa <https://api.inaturalist.org/v1/docs/#!/Taxa/get_taxa>`
    """

    observation_id: int = field(default=None, doc='Associated observation ID')
    original_dimensions: Tuple[int, int] = field(
        converter=format_dimensions, default=(0, 0), doc='Dimensions of original image'
    )
    url: str = field(default=None, doc='Image URL; see properties for URLs of specific image sizes')
    user_id: int = field(default=None, doc='Associated user ID')
    uuid: str = field(default=None)
    _url_format: str = field(init=False, repr=False, default=None)

    # Unused attributes
    # flags: List = field(factory=list)

    def __attrs_post_init__(self):
        # If there's no URL, make a guess based on ID and license:
        self.url = self.url or (
            f'{PHOTO_CC_BASE_URL}/{self.id}/original.jpg'
            if self.has_cc_license
            else f'{PHOTO_BASE_URL}/{self.id}?size=original'
        )

        # Get a URL format string to get different photo sizes. Note: default URL may be any size.
        for size in PHOTO_SIZES:
            if f'{size}.' in self.url:
                self._url_format = self.url.replace(size, '{size}')

    @classmethod
    def from_json(cls, value: JsonResponse, **kwargs) -> 'Photo':
        """Flatten out potentially nested photo field before initializing"""
        if 'photo' in value:
            value.update(value.pop('photo'))
        return super(Photo, cls).from_json(value, **kwargs)

    @property
    def dimensions_str(self) -> str:
        """Dimensions as a string, formatted as ``{width}x{height}``"""
        return f'{self.original_dimensions[0]}x{self.original_dimensions[1]}'

    @property
    def info_url(self) -> str:
        """Photo info URL on iNaturalist.org"""
        return f'{PHOTO_INFO_BASE_URL}/{self.id}'

    @property
    def large_url(self) -> Optional[str]:
        """Image URL (large size)"""
        return self.url_size('large')

    @property
    def medium_url(self) -> Optional[str]:
        """Image URL (medium size)"""
        return self.url_size('medium')

    @property
    def mimetype(self) -> str:
        """MIME type of the image"""
        return f'image/{self.ext.replace("jpg", "jpeg")}'

    @property
    def original_url(self) -> Optional[str]:
        """Image URL (original size)"""
        return self.url_size('original')

    @property
    def small_url(self) -> Optional[str]:
        """Image URL (small size)"""
        return self.url_size('small')

    @property
    def square_url(self) -> Optional[str]:
        """Image URL (thumbnail size)"""
        return self.url_size('square')

    @property
    def thumbnail_url(self) -> Optional[str]:
        """Image URL (thumbnail size)"""
        return self.url_size('square')

    def url_size(self, size: str) -> Optional[str]:
        size = size.replace('thumbnail', 'square').replace('thumb', 'square')
        if not self._url_format or size not in PHOTO_SIZES:
            return None
        return self._url_format.format(size=size)

    def open(self, size: str = 'large') -> BinaryIO:
        """Download the image and return as a file-like object"""
        url = self.url_size(size) or self.url
        return BytesIO(requests.get(url, stream=True).raw.read())

    def show(self, size: str = 'large'):
        """Display the image with the system's default image viewer. Requires ``pillow``."""
        from PIL import Image

        img = Image.open(self.open(size=size))
        img.show()

    @property
    def _row(self) -> TableRow:
        return {
            'ID': self.id,
            'License': self.license_code,
            'Dimensions': self.dimensions_str,
            'URL': self.original_url,
        }


@define_model
class IconPhoto(Photo):
    """:fa:`camera` Class used for displaying an iconic taxon in place of a taxon photo"""

    iconic_taxon_name: str = field(default=None, doc='Iconic taxon name')

    def __attrs_post_init__(self):
        self._url_format = self.url.replace('.png', '-{size}px.png')
        self.url = self.medium_url

    @classmethod
    def from_iconic_taxon(cls, iconic_taxon_name: str):
        url = f'{ICONIC_TAXA_BASE_URL}/{iconic_taxon_name.lower()}.png'
        return cls(url=url)  # type: ignore  # A weird false positive as of mypy 0.950

    @property
    def icon_url(self) -> Optional[str]:
        """Image URL (32px icon size)"""
        return self.url_size('icon')

    def url_size(self, size: str) -> str:
        size = size.replace('thumbnail', 'square').replace('thumb', 'square')
        return self._url_format.format(size=ICON_SIZES.get(size, 'square'))

    @property
    def _str_attrs(self) -> List[str]:
        return ['iconic_taxon_name', 'url']


@define_model
class Sound(BaseMedia):
    """:fa:`media-playback` An observation sound, based on the schema of sounds from:

    * `GET /observations <https://api.inaturalist.org/v1/docs/#!/Observations/get_observations>`_
    * `GET /taxa <https://api.inaturalist.org/v1/docs/#!/Taxa/get_taxa>`
    """

    file_content_type: str = field(default=None)
    file_url: str = field(default=None)
    native_sound_id: str = field(default=None)
    secret_token: str = field(default=None)
    subtype: str = field(default=None)

    # Unused attributes
    # flags: List = field(factory=list)
    # play_local: bool = field(default=None)

    @classmethod
    def from_json(cls, value: JsonResponse, **kwargs) -> 'Sound':
        """Flatten out potentially nested sound field before initializing"""
        if 'sound' in value:
            value.update(value.pop('sound'))
        return super(Sound, cls).from_json(value, **kwargs)

    # Aliases
    @property
    def mimetype(self) -> str:
        return self.file_content_type

    @property
    def url(self) -> str:
        return self.file_url

    @property
    def _row(self) -> TableRow:
        return {
            'ID': self.id,
            'License': self.license_code,
            'URL': self.url,
        }



from typing import TYPE_CHECKING

from pyinaturalist.paginator import Paginator

if TYPE_CHECKING:
    from pyinaturalist.client import iNatClient


class BaseController:
    """Base class for controllers. A controller provides an interface for all requests related to a
    single resource type. This is mainly for the purpose of splitting up :py:class:`.iNatClient`
    methods by resource type, which keeps code, documentation, and usage more organized and easier
    to navigate.
    """

    def __init__(self, client: 'iNatClient'):
        self.client = client

    def from_ids(self, *object_ids, **params) -> Paginator:
        """Get records by ID"""
        raise NotImplementedError



from requests import HTTPError


class AuthenticationError(HTTPError):
    pass


class ObservationNotFound(HTTPError):
    pass


class TaxonNotFound(HTTPError):
    pass


class TooManyRequests(HTTPError):
    """Error raised for either a 429 response, or pre-emptively before we reach the rate limit"""



from functools import update_wrapper
from inspect import signature
from typing import Any, Callable, Dict, List, Optional, Type

from attr import Attribute, Factory

from pyinaturalist.models import BaseModel

FIELD_DEFAULTS = {
    'default': None,
    'validator': None,
    'repr': True,
    'cmp': None,
    'hash': None,
    'init': False,
    'inherited': False,
}


class LazyProperty(property):
    """A lazy-initialized/cached descriptor for ``attrs`` classes. Its purpose is similar to
    ``@functools.cached_property``, but it works for slotted classes by not relying on ``__dict__``.

    Currently this is used to lazy-load nested model objects for better performance.

    How it's used:

    1. Define an ``attrs`` class
    2. Use ``add_lazy_attrs`` as a field_transformer for the class, which adds a temporary attributes
    3. Define a ``LazyProperty`` class attribute with a converter

    How it works:

    1. During attrs init, the temporary attribute is set containing a raw dict
    2. When ``foo`` is first accessed, it runs the converter on the temp attribute
    3. When ``foo`` is accessed again, the previously converted temp attribute will be returned

    Example::

        # Just pretend these are expensive conversion functions
        def converter_func(value) -> str:
            return str(value)

        def list_converter_func(value) -> List:
            return [value]

        @attrs.define(field_transformer=add_lazy_attrs)
        class MyModel(BaseModel):
            str_field = LazyProperty(converter_func)
            list_field = LazyProperty(list_converter_func)

            # Auto-generated temp attributes will look like:
            # _str_field = field(default=None)
            # _list_field = field(factory=list)

    """

    def __init__(
        self,
        converter: Callable,
        name: Optional[str] = None,
        doc: Optional[str] = None,
        type: Type = BaseModel,
        **converter_kwargs,
    ):
        update_wrapper(self, converter)  # type: ignore
        self.converter = converter
        self.converter_kwargs = converter_kwargs
        self.default: Any = None
        self.type = type
        self.__doc__ = doc
        self.__set_name__(None, name)

        # Use either a list factory or default value, depending on the converter's return type
        if _returns_list(converter):
            self.default = Factory(list)

    def __get__(self, obj, cls):
        """When accessing the value, convert it if it hasn't already been, and cache the converted
        value for subsequent calls.
        """
        if obj is None:
            return self

        value = getattr(obj, self.temp_attr)
        if value and not _is_model_object_or_list(value):
            value = self.converter(value, **self.converter_kwargs)
            setattr(obj, self.temp_attr, value)

        return value

    def __set__(self, obj, raw_value):
        """When initially setting the value, store raw data in the temp attribute"""
        setattr(obj, self.temp_attr, raw_value)

    def __set_name__(self, owner, name):
        """Set the name of both the LazyProperty attribute and its temp attribute"""
        self.__name__ = name
        self.temp_attr = f'_{name}'

    def get_lazy_attr(self) -> Attribute:
        """Get a temp attribute to be used by this LazyProperty instance"""
        return make_attribute(self.temp_attr, init=True, repr=False, default=self.default)


def add_lazy_attrs(cls, fields):
    """A field transformer to do some post-processing on a model class while it's being created.
    For each :py:class:`.LazyProperty` on a model class, this adds a corresponding ``attr.field``
    in which to temporarily store a raw dict value that will later be converted into a model object.
    """
    lazy_properties = [p for p in cls.__dict__.values() if isinstance(p, LazyProperty)]
    return list(fields) + [p.get_lazy_attr() for p in lazy_properties]


def get_lazy_properties(cls: Type[BaseModel]) -> Dict[str, LazyProperty]:
    return {k: v for k, v in cls.__dict__.items() if isinstance(v, LazyProperty)}


def make_attribute(name, **kwargs):
    kwargs = {**FIELD_DEFAULTS, **kwargs}
    return Attribute(name=name, **kwargs)


# TODO: Make this more generic by looking for converter function return type instead of BaseModel?
def _is_model_object_or_list(value):
    try:
        return isinstance(value, BaseModel) or isinstance(value[0], BaseModel)
    except (AttributeError, KeyError, TypeError):
        return False


def _returns_list(func: Callable) -> bool:
    """Determine if a function is annotated with a List return type"""
    return_type = signature(func).return_annotation
    return getattr(return_type, '__origin__', None) in (list, List)



from typing import Any, Union

from pyinaturalist.constants import API_V1, JsonResponse
from pyinaturalist.docs import document_request_params
from pyinaturalist.docs import templates as docs
from pyinaturalist.session import delete, post


@document_request_params(docs._ofvs, docs._access_token)
def set_observation_field(
    observation_id: int, observation_field_id: int, value: Any, **params
) -> JsonResponse:
    """Create or update an observation field value on an observation

    .. rubric:: Notes

    * :fa:`lock` :ref:`Requires authentication <auth>`
    * API reference: :v1:`POST /observation_field_values/{id} <post_observation_field_values>`
    * To find an ``observation_field_id``, either user :py:func:`.get_observation_fields` or
      `search observation fields on iNaturalist <https://www.inaturalist.org/observation_fields>`_

    Example:
        >>> # First find an observation field by name, if the ID is unknown:
        >>> response = get_observation_fields('vespawatch_id')
        >>> observation_field_id = response[0]['id']

        >>> set_observation_field(
        ...     7345179,
        ...     observation_field_id,
        ...     value=250,
        ...     access_token=token,
        ... )

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/post_put_observation_field_value.json
                :language: javascript

    Returns:
        The newly updated field value record
    """
    body = {
        'observation_id': observation_id,
        'observation_field_id': observation_field_id,
        'value': value,
    }
    response = post(f'{API_V1}/observation_field_values', json=body, **params)
    return response.json()


def delete_observation_field(observation_field_value_id: Union[int, str], **params) -> JsonResponse:
    """Delete an observation field value from an observation

    .. rubric:: Notes

    * :fa:`lock` :ref:`Requires authentication <auth>`
    * API reference: :v1:`DELETE /observation_field_values/{id} <delete_observation_field_values_id>`

    Example:
        >>> # Observation field value IDs can be found on observation records:
        >>> response = get_observation(70963477)
        >>> ofv_ids = [ofv['id'] for ofv in response['ofvs']]

        >>> for ofv_id in ofv_ids:
        ...     delete_observation_field(ofv_id)


    Args:
        observation_field_value_id: ID or UUID of the observation field value to delete
        access_token: An access token required for user authentication, as returned by :py:func:`.get_access_token()`
    """
    response = delete(f'{API_V1}/observation_field_values/{observation_field_value_id}', **params)
    return response.json()
